__description__ = "a module that holds all the helper functions for exectuables. This will be a bit of a mess, but it will be useful for high-level scheduling"
__author__ = "Reed Essick (reed.essick@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#---------------------------------------------------------------------------------------------------

from collections import defaultdict
import multiprocessing as mp
import os

import numpy as np

### dependencies not in standard Python libraries
from ligo import segments as ligo_segments

from idq import calibration
from idq import classifiers
from idq import condor
from idq import configparser
from idq import exceptions
from idq import factories
from idq import io
from idq import logs
from idq import names
from idq import plots
from idq import reports
from idq import utils

#---------------------------------------------------------------------------------------------------

DEFAULT_BATCH_NUM_BINS = 1
DEFAULT_BATCH_WORKFLOW = 'block'
DEFAULT_BATCH_LOG_LEVEL = logs.DEFAULT_LOG_LEVEL
DEFAULT_INITIAL_LOOKBACK = 3600 # one hour
DEFAULT_NUM_SEGS_PER_BIN = 1

#---------------------------------------------------------------------------------------------------
### batch processes

def condor_batch(gps_start, gps_end, batchdir, nickname, verbose=False):
    '''
    run the training process
    '''
    ### FIXME: record progress in a logger somewhere!

    ### read in the package prepared by idq-train or idq-straming_train
    diskreporter = io.PickleReporter(batchdir, gps_start, gps_end)
    args, kwargs = diskreporter.retrieve(names.nickname2condorname(nickname))

    ### execute batch
    _batch(gps_start, gps_end, *args, **kwargs)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

def _batch(
        gps_start,
        gps_end,
        segments,
        exclude_in_training,
        exclude_in_else,
        tempdir,
        temppath,
        finaldir,
        finalpath,
        tag,
        trainreporter_kwargs,
        evaluatereporter_kwargs,
        calibratereporter_kwargs,
        nicknames,
        verbose=False,
        batch_logger=None,
        skip_timeseries=False,
        skip_report=False,
    ):
    """
    a helper function for forked workflow via multiprocessing
    """
    batch_logger_exists = batch_logger is not None

    ### set up reporters
    reporter_factory = factories.ReporterFactory()

    logdir = names.tag2logdir(tag, rootdir=tempdir)
    trainreporter = reporter_factory(names.tag2traindir(tag, rootdir=tempdir), gps_start, gps_end, **trainreporter_kwargs)
    evaluatereporter = reporter_factory(names.tag2evaluatedir(tag, rootdir=tempdir), gps_start, gps_end, **evaluatereporter_kwargs)
    calibratereporter = reporter_factory(names.tag2calibratedir(tag, rootdir=tempdir), gps_start, gps_end, **calibratereporter_kwargs)

    final_logdir = names.tag2logdir(tag, rootdir=finaldir)
    final_trainreporter = reporter_factory(names.tag2traindir(tag, rootdir=finaldir), gps_start, gps_end, **trainreporter_kwargs)
    final_evaluatereporter = reporter_factory(names.tag2evaluatedir(tag, rootdir=finaldir), gps_start, gps_end, **evaluatereporter_kwargs)
    final_calibratereporter = reporter_factory(names.tag2calibratedir(tag, rootdir=finaldir), gps_start, gps_end, **calibratereporter_kwargs)

    ### run train job, dropping results in the temporary directory
    if batch_logger_exists:
        batch_logger.info('--- TRAIN : %s ---'%logs.get_logpath(logdir, names.tag2train_logname(tag)))
    train(gps_start, gps_end, temppath, verbose=verbose, exclude=exclude_in_training)

    ### run evaluate job, dropping results in temporary directory
    if batch_logger_exists:
        batch_logger.info('--- EVALUATE : %s ---'%logs.get_logpath(logdir, names.tag2evaluate_logname(tag)))
    evaluate(gps_start, gps_end, temppath, verbose=verbose, exclude=exclude_in_else)

    ### run calibrate job, dropping results in temporary directory
    if batch_logger_exists:
        batch_logger.info('--- CALIBRATE : %s ---'%logs.get_logpath(logdir, names.tag2calibrate_logname(tag)))
    calibrate(gps_start, gps_end, temppath, verbose=verbose)

    ### copy models, quivers, calibration maps into final locations
    # copy the resulting models into the appropriate places within evaluate-dir
    if batch_logger_exists:
        batch_logger.info('--- COPYING RESULTS ---')
    segments = utils.remove_segments(segments, exclude_in_else) ### the segments into which we need to copy the models, etc.
                                                                ### we pass in "segments" as an argument to avoid re-computing scisegs from segdb
    for nickname in nicknames:
        train_topic = names.nickname2train_topic(nickname)
        evaluate_topic = names.nickname2evaluate_topic(nickname)
        calibrate_topic = names.nickname2calibrate_topic(nickname)
        
        model = trainreporter.retrieve(train_topic) ### only do I/O once
        quiver = evaluatereporter.retrieve(evaluate_topic)
        calib = calibratereporter.retrieve(calibrate_topic)

        for start, end in segments: # iterate through segments on which we want to have final evaluations
            # copy model
            if batch_logger_exists:
                batch_logger.info('copying trained model for %s within [%.3f, %.3f) to final location'%(nickname, start, end))
            model.hash = names.start_end2hash(start, end) # update hash to point at new model
            final_trainreporter.start = start
            final_trainreporter.end = end
            final_trainreporter.report(train_topic, model, preferred=True)

            # copy calibration maps
            if batch_logger_exists:
                batch_logger.info('copying calibration map for %s within [%.3f, %.3f) to final location'%(nickname, start, end))
            calib.hash = names.start_end2hash(start, end) # update hash to point at new calib map
            final_calibratereporter.start = start
            final_calibratereporter.end = end
            final_calibratereporter.report(calibrate_topic, calib, preferred=True)

            # copy (subsets) of quivers
            if batch_logger_exists:
                batch_logger.info('copying (subset) of evaluated quiver for %s within [%.3f, %.3f) to final location'%(nickname, start, end))
            nsegments = utils.remove_segments(segments, ligo_segments.segmentlist([ligo_segments.segment(start, end)])) ### the segments that we want to keep in the quiver
            subset = quiver.filter(segs=nsegments) ### removes and returns the subset that does not fall within nsegment
            subset.hash = names.start_end2hash(start, end) # update hash to point at new model
            subset.start = start # update this to reflect just the one segment
            subset.end = end

            final_evaluatereporter.start = start ### report
            final_evaluatereporter.end = end
            final_evaluatereporter.report(evaluate_topic, subset, preferred=True)

    ### run timeseries job over each 1-bin stretch separately
    if skip_timeseries:
        if batch_logger_exists:
            batch_logger.info('--- TIMESERIES : skipping ---')
    else:
        if batch_logger_exists:
            batch_logger.info('--- TIMESERIES : %s ---'%logs.get_logpath(logdir, names.tag2timeseries_logname(tag)))
        for start, end in segments: ### run a separate timeseries job over each segment, which will pick up the model and calibmap copied for that segment
            timeseries(start, end, finalpath, verbose=verbose, exclude=[])

    ### run report job over each 1-bin stretch separately
    if skip_report:
        if batch_logger_exists:
            batch_logger.info('--- REPORT : skipping ---')
    else:
        if batch_logger_exists:
            batch_logger.info('--- REPORT : %s ---'%logs.get_logpath(logdir, names.tag2report_logname(tag)))
        report(gps_start, gps_end, finalpath, verbose=verbose, skip_timeseries=skip_timeseries) ### just do this for EVERYTHING once instead of splitting by segments

def batch(
        gps_start,
        gps_end,
        config_path,
        verbose=False,
        num_bins=DEFAULT_BATCH_NUM_BINS,
        batch_workflow=DEFAULT_BATCH_WORKFLOW,
        batch_log_level=DEFAULT_BATCH_LOG_LEVEL,
        skip_timeseries=False,
        skip_report=False,
        persist=False,
        monitoring_cadence=utils.DEFAULT_MONITORING_CADENCE,
        causal=False,
        num_segs_per_bin=DEFAULT_NUM_SEGS_PER_BIN,
        initial_lookback=DEFAULT_INITIAL_LOOKBACK,
    ):
    '''
    launch and manage all batch processes
    should manage round-robin training/evaluation automatically so users really just have to "point and click"
    '''
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    #---------------------------------------------
    ### set up logging
    #---------------------------------------------
    tag = config.get('general', 'tag')
    rootdir = config.get('general', 'rootdir')
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2batch_logname(tag),
        log_level=batch_log_level,
        rootdir=logdir,
        verbose=verbose
    )
    logger.info( "using config : %s"%config_path )

    #---------------------------------------------
    ### pull out a things that will be used throughout all batch runs
    #---------------------------------------------

    trainreporter_kwargs = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter_kwargs = configparser.add_missing_kwargs(trainreporter_kwargs, group=names.tag2group(tag, 'train'))

    evaluatereporter_kwargs = configparser.evalkwargs(**dict(config.items('evaluate reporting')))
    evaluatereporter_kwargs = configparser.add_missing_kwargs(evaluatereporter_kwargs, group=names.tag2group(tag, 'evaluate'))

    calibratereporter_kwargs = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter_kwargs = configparser.add_missing_kwargs(calibratereporter_kwargs, group=names.tag2group(tag, 'calibrate'))

    batchdir = names.tag2batchdir(tag, rootdir=rootdir)
    nicknames = config.get('general', 'classifiers').split()

    exclude_in_train = []
    exclude_in_else = []
    tempdirs = []
    temppaths = []

    #---------------------------------------------
    ### use num_bins to partition time into separate stretches
    #---------------------------------------------

    if num_bins > 1: ### automatically divide up segments for cross validation
        if causal:
            logger.info('causal round-robin training with %d bins within [%.3f, %.3f)'%(num_bins, gps_start, gps_end))
        else:
            logger.info('round-robin training with %d bins and %d segments per bin within [%.3f, %.3f)'%(num_bins, num_segs_per_bin, gps_start, gps_end))

        # we check to make sure that all the segdb decisions are consistent between train, evaluate, and timeseries jobs
        train_ignore_segdb = config.has_option('train data discovery', 'ignore_segdb') and config.getboolean('train data discovery','ignore_segdb')
        evaluate_ignore_segdb = config.has_option('evaluate data discovery', 'ignore_segdb') and config.getboolean('evaluate data discovery','ignore_segdb')
        ignore_segdb = [train_ignore_segdb, evaluate_ignore_segdb]
        if not skip_timeseries:
            ignore_segdb.append(config.has_option('timeseries data discovery', 'ignore_segdb') and config.getboolean('timeseries data discovery','ignore_segdb'))
        if not skip_report:
            ignore_segdb.append(config.has_option('report data discovery', 'ignore_segdb') and config.getboolean('report data discovery','ignore_segdb'))

        if all(ignore_segdb) != any(ignore_segdb):
            raise exceptions.ConfigurationInconsistency('we must either supply the same ignore_segdb option within all data discovery sections for batch jobs with num_bins>1')

        ### grab segments
        if train_ignore_segdb:
            logger.info( "ignoring segdb" )

            segs = ligo_segments.segmentlist([ligo_segments.segment(gps_start, gps_end)])
            if causal:
                initial_segs = ligo_segments.segmentlist([ligo_segments.segment(gps_start-initial_lookback, gps_start)])
            else:
                initial_segs = ligo_segments.segmentlist([])

        else:
            segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
            segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
            segdb_url = config.get('segments', 'segdb_url')

            logger.info( "querying %s within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%\
                (segdb_url, gps_start, gps_end, ','.join(segdb_intersect), ','.join(segdb_exclude))
            )

            segs = utils.segdb2segs(
                gps_start,
                gps_end,
                intersect=segdb_intersect,
                exclude=segdb_exclude,
                segdb_url=segdb_url,
                logger=logger,
            )
            if causal:
                logger.info( "querying %s for intial lookback within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%\
                    (segdb_url, gps_start-initial_lookback, gps_start, ','.join(segdb_intersect), ','.join(segdb_exclude))
                )
                initial_segs = utils.segdb2segs(
                    gps_start-initial_lookback,
                    gps_start,
                    intersect=segdb_intersect,
                    exclude=segdb_exclude,
                    segdb_url=segdb_url,
                    logger=logger,
                )
            else:
                initial_segs = ligo_segments.segmentlist([])

        segdb_exclude = utils.fix_segments(ligo_segments.segmentlist(segdb_exclude))
        logger.info('excluding segments : '+str(segdb_exclude))

        segs = utils.remove_segments(segs, segdb_exclude) ### remove any times specified on the command line
        initial_segs = utils.remove_segments(initial_segs, segdb_exclude)

        livetime = utils.livetime(segs)
        if causal:
            initial_livetime = utils.livetime(initial_segs)
            logger.info('retained %.3f+%.3f sec of livetime'%(initial_livetime, livetime))
            livetime += initial_livetime
        else:
            logger.info('retained %.3f sec of livetime'%livetime)

        ### dsplit up the data into bins
        logger.info('dividing bins equally based on livetime:')
        if causal:
            folds = utils.segments_causal_kfold(segs, num_bins, initial_segs=initial_segs)
            gps_start -= initial_lookback ### include this in the call to _batch
        else:
            folds = utils.segments_checkerboard(segs, num_bins, num_segs_per_bin)
        segs = utils.segments_union(segs, initial_segs) ### NOTE: we only take the union AFTER we divide up the analysis time into folds
                                                        ### this is done to simplify the logic in the following loop, which defines segments
                                                        ### that should be excluded based on "segs", which therefore must include "initial_segs"
                                                        ### for causal training

        # assign data into bins
        for i, (trainsegs, elsesegs) in enumerate(folds, 1):
            ### format batch, exclude segments
            ### we do this backflip to make sure we only pick up the right parts for causal batch jobs
            exclude_in_train.append( utils.remove_segments(segs, trainsegs) ) ### exclude everything that is not in trainsegs
            exclude_in_else.append( utils.remove_segments(segs, elsesegs) )   ### exclude everything that is not in elsesegs
            logger.info('    bin %d -- train livetime: %.3f sec, else livetime: %.3f sec'%\
                (i, livetime-utils.livetime(exclude_in_train[-1]), livetime-utils.livetime(exclude_in_else[-1])))

            ### set up temporary paths
            tempdir = os.path.join(batchdir, 'train_%d-%d-%d'%(i, gps_start, gps_end))
            if not os.path.exists(tempdir):
                os.makedirs(tempdir)
            temppath = os.path.join(tempdir, 'train_%d-%d-%d.ini'%(i, gps_start, gps_end))

            logger.info('writing temporary config to: '+temppath)
            config.set('general', 'rootdir', tempdir)
            with open(temppath, 'w') as file_obj:
                config.write(file_obj)

            # record stuff
            tempdirs.append(tempdir)
            temppaths.append(temppath)

    else: # num_bins == 1, so train and evaluate on the same stretch
        segment = ligo_segments.segment(gps_start, gps_end)
        segs = ligo_segments.segmentlist([segment])
        logger.info('training with 1 bin within [%.3f, %.3f)'%segment)
        logger.warn('You are training and evaluating with a single bin. Be careful! This can produce artificially inflated performance estimates.')

        # record stuff
        exclude_in_train.append([])
        exclude_in_else.append([])
        tempdirs.append(rootdir)
        temppaths.append(config_path)

    #---

    if batch_workflow=='block':
        logger.info('processing segments sequentially')

        for eit, eie, tempdir, temppath in zip(exclude_in_train, exclude_in_else, tempdirs, temppaths):
            _batch(
                gps_start,
                gps_end,
                segs,
                eit,
                eie,
                tempdir,
                temppath,
                rootdir,
                config_path,
                tag,
                trainreporter_kwargs,
                evaluatereporter_kwargs,
                calibratereporter_kwargs,
                nicknames,
                verbose=verbose,
                batch_logger=logger,
                skip_timeseries=skip_timeseries,
                skip_report=skip_report,
            )

    #---

    elif batch_workflow=='fork':
        logger.info( 'processing segments in parallel via multiprocessing' )
        workers = []

        try:
            workers = []
            kwargs = {'verbose':verbose, 'skip_timeseries':skip_timeseries, 'skip_report':skip_report}
            for eit, eie, tempdir, temppath in zip(exclude_in_train, exclude_in_else, tempdirs, temppaths):
                args = (
                gps_start,
                gps_end,
                segs,
                eit,
                eie,
                tempdir,
                temppath,
                rootdir,
                config_path,
                tag,
                trainreporter_kwargs,
                evaluatereporter_kwargs,
                calibratereporter_kwargs,
                nicknames,
                )
                proc = mp.Process(target=_batch, args=args, kwargs=kwargs)
                proc.start()
                workers.append( proc )

            ### monitor forked jobs, cleaning up as they finish
            if persist:
                manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
                while workers:
                    proc = workers.pop()
                    if proc.is_alive(): ### still going
                        workers.append( proc ) ### add it back in

                    else: ### process is done
                        if proc.exitcode: ### something went wrong with this process!
                            raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                        proc.terminate() ### make sure we clean up the process

                    manager.wait()

        except Exception as e:
            while workers:
                proc = workers.pop()
                proc.terminate()
            raise e

    #---

    elif batch_workflow=='condor':
        logger.info( 'processing segments in parallel via condor' )

        diskreporter = io.PickleReporter(batchdir, gps_start, gps_end)
        batchnames = []
        kwargs = {'verbose':verbose, 'skip_timeseries':skip_timeseries, 'skip_report':skip_report}
        for ind, (eit, eie, tempdir, temppath) in enumerate(zip(exclude_in_train, exclude_in_else, tempdirs, temppaths)):
            ### prepare package for idq-condor_batch
            args = (\
                gps_start,
                gps_end,
                segs,
                eit,
                eie,
                tempdir,
                temppath,
                rootdir,
                config_path,
                tag,
                trainreporter_kwargs,
                evaluatereporter_kwargs,
                calibratereporter_kwargs,
                nicknames,
            )
            # write to disk
            batchname = names.bin2batchname(ind)
            path = diskreporter.report(names.nickname2condorname(batchname), (args, kwargs))
            logger.info('blob written to: '+path)
            batchnames.append(batchname)

        dag_path = condor.batch_dag(
            gps_start,
            gps_end,
            batchdir,
            batchnames,
            **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
        )
        logger.info( 'dag written to '+dag_path )

        exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
        if exitcode:
            raise RuntimeError('non-zero returncode for dag '+dag_path )

        if persist:
            raise NotImplementedError('we currently cannot monitor condor for completion of batch jobs...')

    #---

    else:
        raise ValueError('batch_workflow=%s not understood'%batch_workflow)

    return None

#-------------------------------------------------

def condor_train(gps_start, gps_end, traindir, nickname, verbose=False):
    '''
    run the training process
    '''
    ### FIXME: record progress in a logger somewhere!

    ### read in the package prepared by idq-train or idq-straming_train
    diskreporter = io.PickleReporter(traindir, gps_start, gps_end)
    blob = diskreporter.retrieve(names.nickname2condorname(nickname))

    ### extract the objects we need
    classifier = blob['classifier']                               ### FIXME don't hard code these names, but look them up in names.py?
    args = diskreporter.retrieve(blob['args_nickname'])
    reporter = diskreporter.retrieve(blob['reporter_nickname'])

    ### execute training
    model = classifier.train(*args)

    ### report the model
    reporter.start, reporter.end = model.start, model.end ### update this to match the model's provenance
    reporter.report(names.nickname2train_topic(classifier.nickname), model, preferred=True)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

def _train_and_report(conn, classifier, reporter, *args):
    """
    a helper function for forked workflow via multiprocessing
    """
    model = classifier.train(*args)
    reporter.start, reporter.end = model.start, model.end ### update this to match the model's provenance
    path = reporter.report(names.nickname2train_topic(classifier.nickname), model, preferred=True) ### report the final result
    conn.send(path)

def train(gps_start, gps_end, config_path, verbose=False, exclude=[]):
    '''
    run the training process
    '''
    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(os.path.realpath(config.get('general','rootdir')))
    traindir = names.tag2traindir(tag, rootdir) ### used to store results
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    target_channel = config.get('samples', 'target_channel')
    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))

    dirty_bounds = configparser.config2bounds(config.get('samples', 'dirty_bounds'))
    dirty_window = config.getfloat('samples', 'dirty_window')

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2train_logname(tag),
        log_level=config.getint('train','log_level'),
        rootdir=logdir,
        verbose=verbose
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'train')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(traindir, gps_start, gps_end)  ### we could use this for everything
    segmentreporter = io.SegmentReporter(                             ### but these handle common data formats
        traindir,
        gps_start,
        gps_end,
    )
    gpstimesreporter = io.GPSTimesReporter(traindir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actuall instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=traindir, **configparser.evalkwargs(**dict(items))))

    ### sort classifiers by how they train (their syntax)
    train_with_quiver = []
    train_wout_quiver = [] ### NOTE: implicitly, this means we train with classifier_data and target_times (support for OVL)
    for classifier in klassifiers:
        if classifier.trains_with_quiver:
            train_with_quiver.append(classifier)
        else:
            train_wout_quiver.append(classifier)

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab segments
    if config.has_option('train data discovery', 'ignore_segdb') and config.getboolean('train data discovery','ignore_segdb'):
        logger.info( "ignoring segdb" )

        segs = ligo_segments.segmentlist([ligo_segments.segment(gps_start, gps_end)])

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')

        logger.info( "querying %s within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%(segdb_url, gps_start, gps_end, ','.join(segdb_intersect), ','.join(segdb_exclude)) )

        segs = utils.segdb2segs(
            gps_start,
            gps_end,
            intersect=segdb_intersect,
            exclude=segdb_exclude,
            segdb_url=segdb_url,
            logger=logger,
         )

    logger.info('excluding segments : '+str(exclude))
    segs = utils.remove_segments(segs, utils.fix_segments(ligo_segments.segmentlist(exclude))) ### remove any times specified on the command line

    logger.info('retained %.3f sec of livetime'%utils.livetime(segs))

    # record segments
    path = segmentreporter.report('segments', segs)
    logger.info( 'segments written to '+path )

    ### grab data
    items = config.items('train data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'train'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)
    time = items.pop('time') ### extract the name of the column we use as time

    classifier_data = classifier_data_factory(
        gps_start,
        gps_end,
        segs=segs,
        **items
    )

    # record classifier_data
    path = picklereporter.report('classifierData', classifier_data)
    logger.info( 'classifier_data written to '+path )

    ### define ``truth'' labels for samples
    #---
    # NOTE: if we switch to continuous weighting of samples instead of discrete labels, this whole block will need to be re-written...
    #---
    # we will always need these for both quiver construction and training with OVL (trains_with_quiver=False)
    logger.info( 'identifying target times' )
    target_times = classifier_data.target_times(time, target_channel, target_bounds)
    logger.info( 'identified %d target_times'%len(target_times) )

    # record target times
    path = gpstimesreporter.report('targetTimes', target_times)
    logger.info( 'target_times written to '+path ) ### FIXME: write to a standard data format!

    # make the quiver
    if train_with_quiver: ### at least one classifier trains with a quiver, so we better make one

        ### only draw from segments that are outside of dirtysegs but within classifier_data.segs
        random_times = classifier_data.random_times(time, target_channel, dirty_bounds, dirty_window, config.getfloat('train', 'random_rate'))
        logger.info( 'identified %d random times'%len(random_times) )

        # record random times
        path = gpstimesreporter.report('randomTimes', random_times) ### FIXME: write to a standard data format!
        logger.info( 'random_times written to '+path )

        ### actually make the quiver
        logger.info( 'building quiver' )
        quiver = factories.QuiverFactory(classifier_data, avg_rate_threshold=avg_rate_threshold).labeled(target_times, random_times)

        # record quiver
        quiverreporter = io.QuiverReporter(traindir, gps_start, gps_end)
        path = quiverreporter.report('quiver', quiver)
        logger.info( 'quiver written to '+path )

    #-----------------------------------------
    ### actually run jobs

    workflow = config.get('train', 'workflow')
    if workflow=='fork':
        workers = [] ### declare this to avoid a race condition when error reporting below

    ### run in series
    if workflow == 'block':
        logger.info( 'training classifiers sequentially' )

        for classifier in train_with_quiver:
            nickname = classifier.nickname
            logger.info( 'training %s with quiver'%nickname )
            model = classifier.train(quiver)
            trainreporter.start, trainreporter.end = model.start, model.end ### update this to match the model's provenance
            logger.info('reporting model')
            path = trainreporter.report(names.nickname2train_topic(nickname), model, preferred=True) ### report the resulting model
            logger.info( 'model for %s written to %s'%(nickname, path) )

        for classifier in train_wout_quiver:
            nickname = classifier.nickname
            logger.info( 'training %s with classifier_data, target_times'%nickname )
            model = classifier.train(classifier_data, target_times)
            trainreporter.start, trainreporter.end = model.start, model.end ### update this to match the model's provenance
            logger.info('reporting model')
            path = trainreporter.report(names.nickname2train_topic(nickname), model, preferred=True) ### report the resulting model
            logger.info( 'model for %s written to %s'%(nickname, path) )

    ### run in parallel on the same node
    elif workflow == 'fork':
        ### FIXME: do we really need this? or can we get away with using the local universe within condor?
        ### FIXME: do we want to specify a maximum number of processes? Currently we just spawn all of them...could be bad...
        logger.info( 'training classifiers in parallel via multiprocessing' )

        try:
            ### set up all the jobs
            for classifier in train_with_quiver:
                nickname = classifier.nickname
                logger.info( 'training %s with quiver'%nickname )
                conn1, conn2 = mp.Pipe()
                args = (conn1, classifier, trainreporter, quiver)
                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                proc.start()
                conn1.close()
                workers.append( (proc, conn2, args) )

            for classifier in train_wout_quiver:
                nickname = classifier.nickname
                logger.info( 'training %s with classifier_data, target_times'%nickname )
                conn1, conn2 = mp.Pipe()
                args = (conn1, classifier, trainreporter, classifier_data, target_times)
                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                proc.start()
                conn1.close()
                workers.append( (proc, conn2, args) )

            ### monitor forked jobs, cleaning up as they finish
            if config.has_option('train', 'monitoring_cadence'):
                monitoring_cadence = config.getfloat('train', 'monitoring_cadence')
            else:
                monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
            manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
            while workers:
                proc, conn2, args = workers.pop()
                if proc.is_alive(): ### still going
                    workers.append( (proc, conn2, args) ) ### add it back in

                else: ### process is done
                    if proc.exitcode: ### something went wrong with this process!
                        raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                    proc.terminate() ### make sure we clean up the process
                    logger.info( 'model for %s written to %s'%(proc.name, conn2.recv()) )
                    conn2.close()

                manager.wait()

        except Exception as e:
            while workers:
                proc, conn2, _ = workers.pop()
                proc.terminate()
                conn2.close()
            raise e

    ### submit to condor
    elif workflow == 'condor':
        logger.info( 'training classifiers in parallel via condor' )

        ### set up blob which we'll write to disk for each classifier
        blob = {
            'classifier':None,
            'args_nickname':None,
            'reporter_nickname':None,
        }

        ### write reporter to disk
        blob['reporter_nickname'] = 'reporter' ### FIXME: don't hard-code nickname? look it up in names.py?
        path = picklereporter.report(blob['reporter_nickname'], trainreporter)
        logger.info( 'reporter written to '+path )

        ### iterate through classifiers, writing a blob for each to disk
        if train_with_quiver:
            ### write quiver to disk
            blob['args_nickname'] = 'trainWithQuiver' ### FIXME: don't hard-code nickname? look it up in names.py?
            path = picklereporter.report(blob['args_nickname'], [quiver])
            logger.info( 'quiver written to '+path )

            ### write blobs for each classifier to disk
            for classifier in train_with_quiver:
                nickname = classifier.nickname
                blob['classifier'] = classifier
                path = picklereporter.report(names.nickname2condorname(nickname), blob)
                logger.info( 'blob for %s written to %s'%(nickname, path) )

        if train_wout_quiver: ### write the rest to disk
            ### write classifier_data, target_times to disk
            blob['args_nickname'] = 'trainWoutQuiver' ### FIXME: don't hard-code nickname? look it up in names.py?
            path = picklereporter.report(blob['args_nickname'], [classifier_data, target_times])
            logger.info( 'classifier_data, target_times written to '+path )

            ### write blobs for each classifier to disk
            for classifier in train_wout_quiver:
                nickname = classifier.nickname
                blob['classifier'] = classifier
                path = picklereporter.report(names.nickname2condorname(nickname), blob)
                logger.info( 'blob for %s written to %s'%(nickname, path) )

        ### define sub, dag files
        dag_path = condor.train_dag(
            gps_start,
            gps_end,
            traindir,
            [classifier.nickname for classifier in klassifiers],
            **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
        )
        logger.info( 'dag written to '+dag_path )

        exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
        if exitcode:
            raise RuntimeError('non-zero returncode for dag '+dag_path )

    ### not understood...
    else:
        raise ValueError('workflow=%s not understood'%(workflow) )

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def condor_evaluate(gps_start, gps_end, evaluatedir, nickname, verbose=False):
    '''
    calibrate the classifier predictions based on historical data
    '''
    ### FIXME: record progress in a logger somewhere!

    ### read in the package prepared by idq-evaluate or idq-streaming_evaluate
    diskreporter = io.PickleReporter(evaluatedir, gps_start, gps_end)
    blob = diskreporter.retrieve(names.nickname2condorname(nickname))

    ### extract the objects we need
    classifier = blob['classifier']                              ### FIXME don't hard code these names, but look them up in names.py?
    args = diskreporter.retrieve(blob['args_nickname'])
    reporter = diskreporter.retrieve(blob['reporter_nickname'])

    ### evaluate labeled quiver for determining classifier performance
    quiver = classifier.evaluate(*args)

    ### report
    reporter.start, reporter.end = quiver.start, quiver.end ### manage provanance
    path = reporter.report(names.nickname2evaluate_topic(classifier.nickname), quiver, preferred=True)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

def _evaluate_and_report(conn, classifier, reporter, *args):
    '''
    a helper function for forked workflow via multiprocessing
    '''
    quiver = classifier.evaluate(*args)
    reporter.start, reporter.end = quiver.start, quiver.end ### manage provenance
    path = reporter.report(names.nickname2evaluate_topic(classifier.nickname), quiver, preferred=True) ### report the final result
    conn.send(path)

def evaluate(gps_start, gps_end, config_path, verbose=False, exclude=[], preferred=False):
    '''
    evaluate the data using classifiers
    '''
    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(os.path.realpath(config.get('general','rootdir')))
    evaluatedir = names.tag2evaluatedir(tag, rootdir=rootdir)
    traindir = names.tag2traindir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, evaluatedir, logdir]:
        if not os.path.exists(directory):
            try:
                os.makedirs(directory)
            except OSError: ### directory already exists
                pass

    ### get info for building quivers
    target_channel = config.get('samples', 'target_channel')
    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))

    dirty_bounds = configparser.config2bounds(config.get('samples', 'dirty_bounds'))
    dirty_window = config.getfloat('samples', 'dirty_window')

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2evaluate_logname(tag),
        log_level=config.getint('evaluate','log_level'),
        rootdir=logdir,
        verbose=verbose,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------
    ritems = configparser.evalkwargs(**dict(config.items('evaluate reporting')))
    evaluatereporter = reporter_factory(evaluatedir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'evaluate')))
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'evaluate')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(evaluatedir, gps_start, gps_end)     ### we could use this for everything
    segmentreporter = io.SegmentReporter(                             ### but these handle common data formats
        evaluatedir,
        gps_start,
        gps_end,
    )
    gpstimesreporter = io.GPSTimesReporter(evaluatedir, gps_start, gps_end)
    quiverreporter = io.QuiverReporter(evaluatedir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actuall instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=evaluatedir, **configparser.evalkwargs(**dict(items))))

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab segments
    if config.has_option('evaluate data discovery', 'ignore_segdb') and config.getboolean('evaluate data discovery','ignore_segdb'):
        logger.info( "ignoring segdb" )

        segs = ligo_segments.segmentlist([ligo_segments.segment(gps_start, gps_end)])

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')

        logger.info( "querying %s within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%(segdb_url, gps_start, gps_end, ','.join(segdb_intersect), ','.join(segdb_exclude)) )

        segs = utils.segdb2segs(
            gps_start,
            gps_end,
            intersect=segdb_intersect,
            exclude=segdb_exclude,
            segdb_url=segdb_url,
            logger=logger,
         )

    logger.info('excluding segments : '+str(exclude))
    segs = utils.remove_segments(segs, utils.fix_segments(ligo_segments.segmentlist(exclude))) ### remove any times specified on the command line

    logger.info('retained %.3f sec of livetime'%utils.livetime(segs))

    ### record segments
    path = segmentreporter.report('segments', segs)
    logger.info( 'segments written to '+path )

    ### grab data
    items = config.items('evaluate data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'evaluate'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)
    time = items.pop('time') ### extract the name of the column we use as time
    classifier_data = classifier_data_factory(
        gps_start,
        gps_end,
        segs=segs,
        **items
    )

    # record classifier_data
    path = picklereporter.report('classifierData', classifier_data)
    logger.info( 'classifier_data written to '+path )

    #-----------------------------------------
    ### generate quiver for evaluation
    #-----------------------------------------
    logger.info( 'identifying target times' )
    target_times = classifier_data.target_times(time, target_channel, target_bounds)
    logger.info( 'identified %d target_times'%len(target_times) )

    # record target times
    path = gpstimesreporter.report('targetTimes', target_times)
    logger.info( 'target_times written to '+path ) ### FIXME: write to a standard data format!

    ### draw random times
    random_times = classifier_data.random_times(time, target_channel, dirty_bounds, dirty_window, config.getfloat('evaluate', 'random_rate'))
    logger.info( 'identified %d random times'%len(random_times) )

    # record random times
    path = gpstimesreporter.report('randomTimes', random_times) ### FIXME: write to a standard data format!
    logger.info( 'random_times written to '+path )

    ### actually make the quiver
    logger.info( 'building quiver' )
    quiver = factories.QuiverFactory(classifier_data, avg_rate_threshold=avg_rate_threshold).labeled(target_times, random_times)

    # record quiver
    path = quiverreporter.report('quiver', quiver)
    logger.info( 'quiver written to '+path )

    #-----------------------------------------
    ### retrieve trained models
    for classifier in klassifiers:
        nickname = classifier.nickname
        logger.info( 'retrieving model for '+nickname )
        classifier.model = trainreporter.retrieve(names.nickname2train_topic(nickname), preferred=preferred)

    #-----------------------------------------
    ### actually run jobs

    workflow = config.get('evaluate', 'workflow')
    if workflow=='fork':
        workers = [] ### declare this to avoid a race condition when error reporting below

    if workflow == 'block':
        logger.info( 'evaluating quiver with trained classifiers sequentially' )

        for classifier in klassifiers:
            nickname = classifier.nickname

            ### evaluate quiver and save probabilities to disk
            logger.info( 'evaluating quiver with %s'%nickname )
            quiver = classifier.evaluate(quiver)

            evaluatereporter.start, evaluatereporter.end = quiver.start, quiver.end ### manage provenance
            logger.info('reporting quiver')
            path = evaluatereporter.report(names.nickname2evaluate_topic(nickname), quiver, preferred=True)
            logger.info( 'quiver written to %s'%path )

    elif workflow == 'fork':
        logger.info( 'evaluating quiver in parallel via multiprocessing' )

        try:
            ### set up all the jobs
            for classifier in klassifiers:
                nickname = classifier.nickname
                logger.info( 'evaluating quiver with %s'%nickname )
                conn1, conn2 = mp.Pipe()
                args = (conn1, classifier, evaluatereporter, quiver)
                proc = mp.Process(target=_evaluate_and_report, args=args, name=nickname)
                proc.start()
                conn1.close()
                workers.append( (proc, conn2, args) )

            ### monitor forked jobs, cleaning up as they finish
            if config.has_option('evaluate', 'monitoring_cadence'):
                monitoring_cadence = config.getfloat('evaluate', 'monitoring_cadence')
            else:
                monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
            manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
            while workers:
                proc, conn2, args = workers.pop()
                if proc.is_alive(): ### still going
                    workers.append( (proc, conn2, args) ) ### add it back in

                else: ### process is done
                    if proc.exitcode: ### something went wrong with this process!
                        raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                    proc.terminate() ### make sure we clean up the process
                    logger.info( 'evaluated quiver for %s written to %s'%(proc.name, conn2.recv()) )
                    conn2.close()

                manager.wait()

        except Exception as e:
            while workers:
                proc, conn2, _ = workers.pop(0)
                proc.terminate()
                conn2.close()
            raise e

    elif workflow == 'condor':
        logger.info( 'evaluating quiver in parallel via condor' )

        ### set up blob which we'll write to disk for each classifier
        blob = {
            'classifier':None,
            'args_nickname':None,
            'reporter_nickname':None,
        }

        ### write reporter to disk
        blob['reporter_nickname'] = 'reporter' ### FIXME: don't hard-code nickname? look it up in names.py?
        path = picklereporter.report(blob['reporter_nickname'], evaluatereporter)
        logger.info( 'reporter written to '+path )

        ### iterate through classifiers, writing a blob for each to disk
        blob['args_nickname'] = 'evaluate' ### FIXME: don't hard-code nickname? look it up in names.py?
        path = picklereporter.report(blob['args_nickname'], [quiver])
        logger.info( 'arguments written to '+path )

        ### write blobs for each classifier to disk
        for classifier in klassifiers:
            nickname = classifier.nickname
            blob['classifier'] = classifier
            path = picklereporter.report(names.nickname2condorname(nickname), blob)
            logger.info( 'blob for %s written to %s'%(nickname, path) )

        ### define sub, dag files
        dag_path = condor.evaluate_dag(
            gps_start,
            gps_end,
            evaluatedir,
            [classifier.nickname for classifier in klassifiers],
            **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
        )
        logger.info( 'dag written to '+dag_path )

        exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
        if exitcode:
            raise RuntimeError('non-zero returncode for dag '+dag_path )

    else:
        raise ValueError('workflow=%s not understood'%(workflow) )

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def condor_calibrate(gps_start, gps_end, calibratedir, nickname, verbose=False):
    '''
    calibrate the classifier predictions based on historical data
    '''
    ### FIXME: record progress in a logger somewhere!

    ### read in the package prepared by idq-calibrate
    diskreporter = io.PickleReporter(calibratedir, gps_start, gps_end)
    blob = diskreporter.retrieve(names.nickname2condorname(nickname))

    ### extract the objects we need
    classifier = blob['classifier']                            ### FIXME don't hard code these names, but look them up in names.py?
    args = diskreporter.retrieve(blob['args_nickname'])
    calibrate_kwargs = args[-1]
    args = args[:-1]
    reporter = diskreporter.retrieve(blob['reporter_nickname'])

    ### evaluate quiver and save probabilities to disk
    calibmap = classifier.calibrate(*args, **calibrate_kwargs)

    ### report
    reporter.start, reporter.end = calibmap.start, calibmap.end ### manage provenance
    path = reporter.report(names.nickname2calibrate_topic(nickname), calibmap, preferred=True)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None


def _calibrate_and_report(conn, classifier, reporter, *args):
    '''
    a helper function for forked workflow via multiprocessing
    '''
    calibrate_kwargs = args[-1]
    args = args[:-1]
    calibmap = classifier.calibrate(*args, **calibrate_kwargs)
    reporter.start, reporter.end = calibmap.start, calibmap.end ### manage provenance
    path = reporter.report(names.nickname2calibrate_topic(classifier.nickname), calibmap, preferred=True) ### report the final result
    conn.send(path)

def calibrate(gps_start, gps_end, config_path, verbose=False, preferred=False):
    '''
    calibrate the classifier predictions based on historical data
    '''
    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(os.path.realpath(config.get('general','rootdir')))
    evaluatedir = names.tag2evaluatedir(tag, rootdir=rootdir)
    calibratedir = names.tag2calibratedir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, evaluatedir, calibratedir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2calibrate_logname(tag),
        log_level=config.getint('calibrate','log_level'),
        rootdir=logdir,
        verbose=verbose,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------
    ritems = configparser.evalkwargs(**dict(config.items('evaluate reporting')))
    evaluatereporter = reporter_factory(evaluatedir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'calibrate')))
    ritems = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter = reporter_factory(calibratedir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'calibrate')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(calibratedir, gps_start, gps_end)     ### we could use this for everything
    segmentreporter = io.SegmentReporter(calibratedir, gps_start, gps_end)   ### but these handle common data formats

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actuall instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=calibratedir, **configparser.evalkwargs(**dict(items))))

    #-----------------------------------------
    # extract basic parameters for calibration
    #-----------------------------------------
    calibrate_kwargs = configparser.evalkwargs(**dict(config.items('calibrate')))

    #-----------------------------------------
    ### actually run jobs

    workflow = config.get('calibrate', 'workflow')
    if workflow=='fork':
        workers = [] ### declare this to avoid a race condition when error reporting below

    if workflow == 'block':
        logger.info( 'calibrating classifiers with evaluated quivers sequentially' )

        for classifier in klassifiers:
            nickname = classifier.nickname

            logger.info('retrieving evaluated quiver')
            quiver = evaluatereporter.retrieve(names.nickname2evaluate_topic(nickname), preferred=preferred)

            ### evaluate quiver and save probabilities to disk
            logger.info( 'calibrating evaluations with %s'%nickname )
            calibmap = classifier.calibrate(quiver, **calibrate_kwargs)

            calibratereporter.start, calibratereporter.end = calibmap.start, calibmap.end ### manage provenance
            logger.info('reporting calibration')
            path = calibratereporter.report(names.nickname2calibrate_topic(nickname), calibmap, preferred=True)
            logger.info( 'calibration written to %s'%path )

    elif workflow == 'fork':
        logger.info( 'calibrating in parallel via multiprocessing' )

        try:
            ### set up all the jobs
            for classifier in klassifiers:
                nickname = classifier.nickname
                logger.info('retrieving evaluated quiver')
                quiver = evaluatereporter.retrieve(names.nickname2evaluate_topic(nickname), preferred=preferred)
                logger.info( 'calibrating evaluations with %s'%nickname )
                conn1, conn2 = mp.Pipe()
                args = (conn1, classifier, calibratereporter, quiver, calibrate_kwargs)
                proc = mp.Process(target=_calibrate_and_report, args=args, name=nickname)
                proc.start()
                conn1.close()
                workers.append( (proc, conn2, args) )

            ### monitor forked jobs, cleaning up as they finish
            if config.has_option('calibrate', 'monitoring_cadence'):
                monitoring_cadence = config.getfloat('calibrate', 'monitoring_cadence')
            else:
                monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
            manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
            while workers:
                proc, conn2, args = workers.pop()
                if proc.is_alive(): ### still going
                    workers.append( (proc, conn2, args) ) ### add it back in

                else: ### process is done
                    if proc.exitcode: ### something went wrong with this process!
                        raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                    proc.terminate() ### make sure we clean up the process
                    logger.info( 'calibration for %s written to %s'%(proc.name, conn2.recv()) )
                    conn2.close()

                manager.wait()

        except Exception as e:
            while workers:
                proc, conn2, _ = workers.pop(0)
                proc.terminate()
                conn2.close()
            raise e

    elif workflow == 'condor':
        logger.info( 'calibrating in parallel via condor' )

        ### set up blob which we'll write to disk for each classifier
        blob = {
            'classifier':None,
            'args_nickname':None,
            'reporter_nickname':None,
        }

        ### write reporter to disk
        blob['reporter_nickname'] = 'reporter' ### FIXME: don't hard-code nickname? look it up in names.py?
        path = picklereporter.report(blob['reporter_nickname'], calibratereporter)
        logger.info( 'reporter written to '+path )


        ### write blobs for each classifier to disk
        for classifier in klassifiers:
            nickname = classifier.nickname
            logger.info('retrieving evaluated quiver')
            quiver = evaluatereporter.retrieve(names.nickname2evaluate_topic(nickname))

            ### iterate through classifiers, writing a blob for each to disk
            blob['args_nickname'] = 'calibrate' ### FIXME: don't hard-code nickname? look it up in names.py?
            path = picklereporter.report(blob['args_nickname'], [quiver, calibrate_kwargs])
            logger.info( 'arguments written to '+path )

            blob['classifier'] = classifier
            path = picklereporter.report(names.nickname2condorname(nickname), blob)
            logger.info( 'blob for %s written to %s'%(nickname, path) )

        ### define sub, dag files
        dag_path = condor.calibrate_dag(
            gps_start,
            gps_end,
            calibratedir,
            [classifier.nickname for classifier in klassifiers],
            **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
        )
        logger.info( 'dag written to '+dag_path )

        exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
        if exitcode:
            raise RuntimeError('non-zero returncode for dag '+dag_path )

    else:
        raise ValueError('workflow=%s not understood'%(workflow) )

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def report_timeseries(reporter, nickname, seriesdicts, preferred=False):
    paths = []
    for s in seriesdicts:
        reporter.start = s['t0']
        reporter.end = s['t0'] + s['deltaT']*len(s['data'].values()[0]) ### just grab one series
        paths.append(reporter.report(nickname, s, preferred=preferred))
    return paths

def condor_timeseries(gps_start, gps_end, timeseriesdir, nickname, verbose=False):
    '''
    calibrate the classifier predictions based on historical data
    '''
    ### FIXME: record progress in a logger somewhere!

    ### read in the package prepared by idq-calibrate
    diskreporter = io.PickleReporter(timeseriesdir, gps_start, gps_end)
    blob = diskreporter.retrieve(names.nickname2condorname(nickname))

    ### extract the objects we need
    classifier = blob['classifier']                            ### FIXME don't hard code these names, but look them up in names.py?
    instrument, args, template = diskreporter.retrieve(blob['args_nickname'])
    reporter = diskreporter.retrieve(blob['reporter_nickname'])

    ### generate timeseries
    series = classifier.timeseries(*args)
    ans = calibration.series2dict(series, classifier.model, classifier.calibration_map, template)

    ### report
    report_timeseries(reporter, names.nickname2timeseries_topic(instrument, nickname), ans, preferred=True)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

def _timeseries_and_report(conn, classifier, reporter, instrument, args, template):
    '''
    a helper function for forked workflow via multiprocessing
    '''
    series = classifier.timeseries(*args)
    ans = calibration.series2dict(series, classifier.model, classifier.calibration_map, template)
    paths = report_timeseries(reporter, names.nickname2timeseries_topic(instrument, classifier.nickname), ans, preferred=True)
    conn.send(paths)

def timeseries(gps_start, gps_end, config_path, verbose=False, exclude=[], preferred=False):
    '''
    evaluate the data using classifiers
    '''
    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(os.path.realpath(config.get('general','rootdir')))
    timeseriesdir = names.tag2timeseriesdir(tag, rootdir=rootdir)
    calibratedir = names.tag2calibratedir(tag, rootdir=rootdir)
    traindir = names.tag2traindir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, calibratedir, timeseriesdir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    ### extract information necessary for channel names in timeseries files
    assert config.has_option('general', 'instrument'), 'instrument not specified in config file!'
    instrument = config.get('general', 'instrument')

    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))
    frequency = config.get('timeseries data discovery', 'frequency')
    assert frequency in target_bounds, 'must specify a frequency range (called %s) within target_bounds'%frequency
    freq_min, freq_max = target_bounds[frequency]
    assert isinstance(freq_min, int) and isinstance(freq_max, int), 'frequency bounds must be integers!'

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2timeseries_logname(tag),
        log_level=config.getint('timeseries','log_level'),
        rootdir=logdir,
        verbose=verbose,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))
    ritems = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter = reporter_factory(calibratedir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))
    ritems = configparser.evalkwargs(**dict(config.items('timeseries reporting')))
    timeseriesreporter = reporter_factory(timeseriesdir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(timeseriesdir, gps_start, gps_end)     ### we could use this for everything
    segmentreporter = io.SegmentReporter(                             ### but these handle common data formats
        timeseriesdir,
        gps_start,
        gps_end,
    )

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actuall instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=timeseriesdir, **configparser.evalkwargs(**dict(items))))

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab segments
    if config.has_option('timeseries data discovery', 'ignore_segdb') and config.getboolean('timeseries data discovery','ignore_segdb'):
        logger.info( "ignoring segdb" )

        segs = ligo_segments.segmentlist([ligo_segments.segment(gps_start, gps_end)])

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')

        logger.info( "querying %s within [%.3f, %.3f) for: intersect=%s ; exclude=%s"%(segdb_url, gps_start, gps_end, ','.join(segdb_intersect), ','.join(segdb_exclude)) )

        segs = utils.segdb2segs(
            gps_start,
            gps_end,
            intersect=segdb_intersect,
            exclude=segdb_exclude,
            segdb_url=segdb_url,
            logger=logger,
         )

    logger.info('excluding segments : '+str(exclude))
    segs = utils.remove_segments(segs, utils.fix_segments(ligo_segments.segmentlist(exclude))) ### remove any times specified on the command line

    logger.info('retained %.3f sec of livetime'%utils.livetime(segs))

    ### record segments
    path = segmentreporter.report('segments', segs)
    logger.info( 'segments written to '+path )

    ### grab data
    items = config.items('timeseries data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'timeseries'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)
    classifier_data = classifier_data_factory(
        gps_start,
        gps_end,
        segs=segs,
        **items
    )
    quiver_factory = factories.QuiverFactory(classifier_data, avg_rate_threshold=avg_rate_threshold)

    # record classifier_data
    path = picklereporter.report('classifierData', classifier_data)
    logger.info( 'classifier_data written to '+path )

    ### figure out timeseres sampling rate -> dt
    srate = config.getfloat('timeseries', 'srate')
    logger.info('generating timeseries sampled at %.3f Hz'%srate)
    dt = 1./srate ### we pass dt to classifier.timeseries so we only have to do division once. Saves, like, a handful of cycles

    #-----------------------------------------
    ### load in model and calibration map
    for classifier in klassifiers:
        nickname = classifier.nickname

        ### set model
        logger.info('retrieving %s classifier'%nickname)
        classifier.model = trainreporter.retrieve(names.nickname2train_topic(nickname), preferred=preferred)

        ### set calibration map
        logger.info('retrieving %s calibration_map'%nickname)
        classifier.calibration_map = calibratereporter.retrieve(names.nickname2calibrate_topic(nickname), preferred=preferred)

    #-----------------------------------------
    ### actually run jobs

    workflow = config.get('timeseries', 'workflow')

    if workflow=='fork':
        workers = [] ### declare this to avoid a race condition when error reporting below

    ### split up timeseries generation into strides is specified
    max_stride = config.getint('timeseries data discovery', 'stride') if config.has_option('timeseries data discovery', 'stride') else None
    if max_stride is not None:
        strides = sorted(utils.split_segments_by_stride(gps_start, gps_end, segs, max_stride))
    else:
        strides = [ligo_segments.segment(gps_start, gps_end)]

    logger.info( 'generating timeseries' )
    for stride in strides:

        ### update report times
        timeseriesreporter.start, timeseriesreporter.end = stride

        logger.info('--- timeseries stride: [%.3f, %.3f) ---'%stride )
        stride_segs = utils.segments_intersection(ligo_segments.segmentlist([stride]), segs)
        lvtm = utils.livetime(stride_segs)

        ### create timeseries if stride livetime is nonzero
        if lvtm:

            if workflow == 'block':
                logger.info( 'generating timeseries sequentially' )
                for classifier in klassifiers:
                    nickname = classifier.nickname

                    ### generate series
                    logger.info('generating timeseries for %s'%nickname)
                    series = classifier.timeseries(quiver_factory, dt=dt, segs=stride_segs)

                    ### calibrate
                    logger.info('calibrating timeseries for %s'%nickname)
                    ans = calibration.series2dict(series, classifier.model, classifier.calibration_map, names.channel_name_template(instrument, nickname, freq_min, freq_max))

                    ### report
                    logger.info('reporting probabilities')
                    for path in report_timeseries(timeseriesreporter, names.nickname2timeseries_topic(instrument, nickname), ans, preferred=True):
                        logger.info( 'probabilities written to %s'%path )

            elif workflow == 'fork':
                logger.info( 'generating timeseries in parallel via multiprocessing' )

                try:
                    ### set up all the jobs
                    for classifier in klassifiers:
                        nickname = classifier.nickname
                        logger.info( 'generating timeseries for %s'%nickname )
                        conn1, conn2 = mp.Pipe()
                        args = (conn1, classifier, timeseriesreporter, instrument, (quiver_factory, dt, stride_segs), names.channel_name_template(instrument, nickname, freq_min, freq_max))
                        proc = mp.Process(target=_timeseries_and_report, args=args, name=nickname)
                        proc.start()
                        conn1.close()
                        workers.append( (proc, conn2, args) )

                    ### monitor forked jobs, cleaning up as they finish
                    if config.has_option('timeseries', 'monitoring_cadence'):
                        monitoring_cadence = config.getfloat('timeseries', 'monitoring_cadence')
                    else:
                        monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
                    manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
                    while workers:
                        proc, conn2, args = workers.pop()
                        if proc.is_alive(): ### still going
                            workers.append( (proc, conn2, args) ) ### add it back in

                        else: ### process is done
                            if proc.exitcode: ### something went wrong with this process!
                                raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                            proc.terminate() ### make sure we clean up the process
                            for path in conn2.recv():
                                logger.info( 'timeseries for %s written to %s'%(proc.name, path) )
                            conn2.close()

                        manager.wait()

                except Exception as e:
                    while workers:
                        proc, conn2, _ = workers.pop(0)
                        proc.terminate()
                        conn2.close()
                    raise e

            elif workflow == 'condor':
                logger.info( 'generating timeseries in parallel via condor' )

                ### set up blob which we'll write to disk for each classifier
                blob = {
                    'classifier':None,
                    'args_nickname':None,
                    'reporter_nickname':None,
                }

                ### write reporter to disk
                blob['reporter_nickname'] = 'reporter' ### FIXME: don't hard-code nickname? look it up in names.py?
                path = picklereporter.report(blob['reporter_nickname'], timeseriesreporter)
                logger.info( 'reporter written to '+path )

                ### iterate through classifiers, writing a blob for each to disk
                blob['args_nickname'] = 'timeseries' ### FIXME: don't hard-code nickname? look it up in names.py?
                path = picklereporter.report(blob['args_nickname'], [instrument, (quiver_factory, dt, stride_segs), names.channel_name_template(instrument, nickname, freq_min, freq_max)])
                logger.info( 'arguments written to '+path )

                ### write blobs for each classifier to disk
                for classifier in klassifiers:
                    nickname = classifier.nickname
                    blob['classifier'] = classifier
                    path = picklereporter.report(names.nickname2condorname(nickname), blob)
                    logger.info( 'blob for %s written to %s'%(nickname, path) )

                ### define sub, dag files
                dag_path = condor.timeseries_dag(
                    gps_start,
                    gps_end,
                    timeseriesdir,
                    [classifier.nickname for classifier in klassifiers],
                    **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
                )
                logger.info( 'dag written to '+dag_path )

                exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
                if exitcode:
                    raise RuntimeError('non-zero returncode for dag '+dag_path )

            else:
                raise ValueError('workflow=%s not understood'%(workflow) )

        else:
            logger.info('no identified times, skipping timeseries')

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def report(gps_start, gps_end, config_path, reportdir=None, zoom_start=None, zoom_end=None, verbose=False, **kwargs):
    """
    generate a report of the classifier predictions in this period
    we allow reportdir to be specified directly here in case you want to make a summary of a job running in a directory you don't own
    """
    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')
    rootdir = os.path.abspath(os.path.realpath(config.get('general','rootdir')))

    if reportdir is None:
        reportdir = names.tag2reportdir(tag, rootdir=rootdir)
        logdir = names.tag2logdir(tag, rootdir=rootdir)
    else:
        logdir = reportdir ### write log into the output directory

    nicknames = config.get('general', 'classifiers').split()

    # hard code some defaults here...
    log_level = config.getint('report', 'log_level')

    ignore_segdb = config.getboolean('report', 'ignore_segdb') if config.has_option('report', 'ignore_segdb') else False
    legend = config.getboolean('report', 'legend') if config.has_option('report', 'legend') else True
    annotate_gch = config.getboolean('report', 'annotate_gch') if config.has_option('report', 'annotate_gch') else False

    ### check various locations to see whether to skip timeseries
    if config.has_option('report', 'skip_timeseries'):
        skip_timeseries = config.getboolean('report', 'skip_timeseries')
    elif 'skip_timeseries' in kwargs:
        skip_timeseries = kwargs['skip_timeseries']
    else:
        skip_timeseries = False

    #-----------------------------------------
    ### set up object and generate report
    #-----------------------------------------

    report_obj = reports.Report(config_path, gps_start, gps_end, t0=gps_start, zoom_start=zoom_start, zoom_end=zoom_end)
    report_obj.report(
        nicknames,
        reportdir=reportdir,
        logdir=logdir,
        ignore_segdb=ignore_segdb,
        skip_timeseries=skip_timeseries,
        annotate_gch_gps=annotate_gch,
        legend=legend,
        verbose=verbose,
        log_level=log_level,
    )
